import os
import re
from datetime import datetime
from .chat_message import EveChatLogReader
from .exceptions import InvalidChatDirectory, InvalidCallable, ObserverAlreadyAdded
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

def get_timestamp_from_file_name(file_name):
    chat_file_timestamp_parser = re.compile('.*?_(\d{8}_\d{6})\.txt$')
    matches = chat_file_timestamp_parser.match(file_name)

    if matches:
        return datetime.strptime(matches.group(1), "%Y%m%d_%H%M%S")

    return None


def get_chat_from_file_name(file_name):
    chat_file_name_parser = re.compile('^(.+?)_\d+_\d+\.')
    matches = chat_file_name_parser.match(file_name)

    if matches:
        return matches.group(1)

    return None


def get_existing_logs(path):
    if not os.path.exists(path):
        raise InvalidChatDirectory(path, "The path '{0}' does not exist.".format(path))

    if not os.path.isdir(path):
        raise InvalidChatDirectory(path,
                                   "The path '{0}' does not point to a directory.".format(path))

    existing_chat_logs = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]

    existing_chats = {}

    for file_name in existing_chat_logs:
        chat_name = get_chat_from_file_name(file_name)
        timestamp = get_timestamp_from_file_name(file_name)

        if chat_name in existing_chats:
            if timestamp > existing_chats[chat_name]['timestamp']:
                existing_chats[chat_name] = {'timestamp': timestamp, 'path': os.path.join(path, file_name)}
        else:
            existing_chats[chat_name] = {'timestamp': timestamp, 'path': os.path.join(path, file_name)}

    return existing_chats


class EveChatLogDirectoryMonitor(object):

    def __init__(self, path):
        if not os.path.exists(path):
            raise InvalidChatDirectory(path, "The path '{0}' does not exist.".format(path))

        if not os.path.isdir(path):
            raise InvalidChatDirectory(path,
                                       "The path '{0}' does not point to a directory.".format(path))
        self.path = path
        self.chats = {}
        self.watchdog_observer = None

        self._add_existing_log_files()
        self._add_file_observer()

    def _add_file_observer(self):
        if self.watchdog_observer:
            raise ObserverAlreadyAdded()

        event_handler = DirChangeEventHandler(self.on_create, self.on_delete)

        self.watchdog_observer = Observer()
        self.watchdog_observer.schedule(event_handler, self.path, recursive=False)

    def _add_existing_log_files(self):
        existing_logs = get_existing_logs(self.path)
        for chat_name, file_info in existing_logs.items():
            self.add_chat_log(chat_name, file_info['path'])

    def read_messages(self, chat_name):
        if chat_name in self.chats:
            return self.chats[chat_name].read_messages()

        return None

    def remove_chat_log(self, chat_name):
        if chat_name in self.chats:
            self.chats[chat_name].destroy()
            del self.chats[chat_name]

    def add_chat_log(self, chat_name, path):
        self.remove_chat_log(chat_name)
        self.chats[chat_name] = EveChatLogReader(path)

    def on_delete(self, event):
        if event.is_directory:
            return

        file_name = os.path.split(event.src_path)[1]

        chat_name = get_chat_from_file_name(file_name)



        self.remove_chat_log(chat_name)

    def on_create(self, event):
        if event.is_directory:
            return

        file_name = os.path.split(event.src_path)[1]
        chat_name = get_chat_from_file_name(file_name)

        self.add_chat_log(chat_name, event.src_path)


class DirChangeEventHandler(FileSystemEventHandler):

    def __init__(self, new_file_callable, file_deleted_callable):

        if not hasattr(new_file_callable, '__call__'):
            raise InvalidCallable("The value passed as '{0}' is not a callable.".format('newFileCallable'))

        if not hasattr(file_deleted_callable, '__call__'):
            raise InvalidCallable("The value passed as '{0}' is not a callable.".format('file_deleted_callable'))

        self.new_file_callable = new_file_callable
        self.file_deleted_callable = file_deleted_callable

    def on_created(self, event):
        self.new_file_callable(event)

    def on_deleted(self, event):
        self.file_deleted_callable(event)





from threading import Thread
from time import sleep
from .chat_directory import EveChatLogDirectoryMonitor
from .chat_message import parse_msg
from .exceptions import InvalidMonitorState

class Monitor:
    """An Eve chat monitor"""

    def __init__(self, chats, path, handler, poll_rate=2):
        self.chats = chats
        self.chat_log_monitor = EveChatLogDirectoryMonitor(path)
        self.handler = handler
        self.is_alive = False
        self.thread = None
        self.poll_rate = poll_rate

    def stop(self):
        if not self.is_alive:
            raise InvalidMonitorState("Monitor not started")
        self.is_alive = False
        self.thread.stop()
        self.thread = None

    def start(self):
        if self.is_alive:
            raise InvalidMonitorState("Monitor already started")
        self.is_alive = True
        self.thread = Thread(target=self.poll)
        self.thread.daemon = True  # thread dies when main thread (only non-daemon thread) exits.
        self.thread.start()

    def _should_poll(self):
        return self.is_alive

    def poll(self):
        while self._should_poll():
            for chat in self.chats:
                messages = self.chat_log_monitor.read_messages(chat)

                for message in messages:
                    self.handler(chat, parse_msg(message))
            sleep(self.poll_rate)







class EveChatMonException(Exception):
    """Base class for all exceptions

        Attributes:
            message -- An explanation of the error
    """
    def __init__(self, message):
        super(EveChatMonException, self).__init__(message)

class InvalidChatDirectory(EveChatMonException):
    """Exception raised when the supplied path to Eve's chat log directory is invalid.

        Attributes:
            path -- The invalid path provided
            message -- An explanation of the error
    """
    def __init__(self, path, message):
        super(InvalidChatDirectory, self).__init__(message)
        self.path = path

class InvalidCallable(EveChatMonException):
    """Exception raised when a supplied argument does not support __call__"""
    def __init__(self, message):
        super(InvalidCallable, self).__init__(message)


class InvalidMonitorState(EveChatMonException):
    def __init__(self, message):
        super(InvalidMonitorState, self).__init__(message)

class ObserverAlreadyAdded(EveChatMonException):
    def __init__(self):
        super(ObserverAlreadyAdded, self).__init__("File monitor already initialized")



import re
from datetime import datetime
from hashlib import md5

def hash(str):
    return md5.hash(message).hexdigest()


def parse_msg(msg):
    line_parser = re.compile('^\s*\[\s(.*?)\s\]\s(.*?)\s>\s(.*?)$', re.DOTALL)
    match = line_parser.match(msg)
    if match:
        timestamp = match.group(1)
        username = match.group(2)
        message = match.group(3)
        message_hash = hash(message)
        timestamp = datetime.strptime(timestamp, "%Y.%m.%d %H:%M:%S")

        parsed_msg = {"timestamp": timestamp,
                      "username": username,
                      "message": message,
                      "line": msg,
                      "hash": message_hash}
        return parsed_msg

    return None

class EveChatLogReader(object):
    chat_line_delimiter = u"\ufeff"

    def __init__(self, path):
        self.path = path
        self.file_handle = open(path, "r", encoding="utf-16-le")
        self.file_handle.read(None)

    def read_messages(self):
        buffer = self.file_handle.read(None)
        clean_messages = []
        if buffer:
            messages = buffer.split(EveChatLogReader.chat_line_delimiter)
            for i in range(len(messages)):
                if messages[i] and messages[i] != '':
                    msg = messages[i].replace(EveChatLogReader.chat_line_delimiter, '')
                    msg = re.sub('\s*$', '', msg)
                    clean_messages.append(msg)

        return clean_messages

    def destroy(self):
        self.file_handle.close()



from setuptools import setup, find_packages

setup(
    name="py-eve-chat-mon",
    version="0.4",
    author="Andrew Martinez",
    author_email="andrew.p.martinez@gmail.com",
    install_requires=[
        "watchdog==0.8.3"
    ],
    description=("A library that focuses on monitoring EVE Online chat logs for messages and doing nothing else. It is meant to be focused and lightweight."),
    license="MIT",
    keywords="EVE chat monitor",
    url='https://github.com/andrewpmartinez/py-eve-chat-mon',
    download_url='https://github.com/andrewpmartinez/py-eve-chat-mon/tarball/0.',
    packages=find_packages(),
    long_description="See github page for full details.",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries'
    ]
)
